@@ -1,10 +1,8 @@
 #pragma once
 
-#define PY_SSIZE_T_CLEAN
 #include <Python.h>
 #include <structmember.h>
 
-#define VK_NO_PROTOTYPES
 #include <vulkan/vulkan_core.h>
 
 #ifdef BUILD_WINDOWS
@@ -61,6 +59,12 @@ struct RenderParameters {
     uint32_t instance_count;
     uint32_t index_count;
     uint32_t indirect_count;
+    uint32_t max_draw_count;
+    VkDeviceSize vertex_buffer_offset;
+    VkDeviceSize instance_buffer_offset;
+    VkDeviceSize index_buffer_offset;
+    VkDeviceSize indirect_buffer_offset;
+    VkDeviceSize count_buffer_offset;
 };
 
 struct ComputeParameters {
@@ -68,6 +72,7 @@ struct ComputeParameters {
     uint32_t x, y, z;
 };
 
+typedef void (* RenderCommand)(RenderPipeline * self, VkCommandBuffer command_buffer);
 typedef void (* Packer)(char * ptr, PyObject ** obj);
 
 struct Format {
@@ -140,6 +145,17 @@ struct ModuleState {
     PyObject * output_str;
 };
 
+struct Extension {
+    VkBool32 acceleration_structure;
+    VkBool32 dedicated_allocation;
+    VkBool32 deferred_host_operations;
+    VkBool32 draw_indirect_count;
+    VkBool32 mesh_shader;
+    VkBool32 pipeline_library;
+    VkBool32 ray_query;
+    VkBool32 ray_tracing_pipeline;
+};
+
 struct Instance {
     PyObject_HEAD
 
@@ -152,13 +168,17 @@ struct Instance {
     VkCommandPool command_pool;
     VkCommandBuffer command_buffer;
 
+    VkPipelineCache pipeline_cache;
     VkDebugUtilsMessengerEXT debug_messenger;
 
+    VkBool32 debug;
+    uint32_t api_version;
     uint32_t queue_family_index;
     uint32_t host_memory_type_index;
     uint32_t device_memory_type_index;
     VkFormat depth_format;
 
+    Extension extension;
     Presenter presenter;
 
     PyObject * task_list;
@@ -171,10 +191,15 @@ struct Instance {
     ModuleState * state;
 
     PFN_vkGetInstanceProcAddr vkGetInstanceProcAddr;
+    PFN_vkEnumerateInstanceVersion vkEnumerateInstanceVersion;
+    PFN_vkEnumerateInstanceLayerProperties vkEnumerateInstanceLayerProperties;
+    PFN_vkEnumerateInstanceExtensionProperties vkEnumerateInstanceExtensionProperties;
     PFN_vkCreateInstance vkCreateInstance;
     PFN_vkDestroyInstance vkDestroyInstance;
     PFN_vkEnumeratePhysicalDevices vkEnumeratePhysicalDevices;
+    PFN_vkEnumerateDeviceExtensionProperties vkEnumerateDeviceExtensionProperties;
     PFN_vkGetPhysicalDeviceProperties vkGetPhysicalDeviceProperties;
+    PFN_vkGetPhysicalDeviceFeatures2 vkGetPhysicalDeviceFeatures2;
     PFN_vkGetPhysicalDeviceMemoryProperties vkGetPhysicalDeviceMemoryProperties;
     PFN_vkGetPhysicalDeviceFeatures vkGetPhysicalDeviceFeatures;
     PFN_vkGetPhysicalDeviceFormatProperties vkGetPhysicalDeviceFormatProperties;
@@ -188,8 +213,8 @@ struct Instance {
     PFN_vkGetImageMemoryRequirements vkGetImageMemoryRequirements;
     PFN_vkCmdCopyImageToBuffer vkCmdCopyImageToBuffer;
     PFN_vkCreateShaderModule vkCreateShaderModule;
+    PFN_vkDestroyShaderModule vkDestroyShaderModule;
     PFN_vkCmdCopyBufferToImage vkCmdCopyBufferToImage;
-    PFN_vkCmdDrawIndirect vkCmdDrawIndirect;
     PFN_vkCreateImageView vkCreateImageView;
     PFN_vkCmdDispatch vkCmdDispatch;
     PFN_vkCmdBindIndexBuffer vkCmdBindIndexBuffer;
@@ -198,7 +223,6 @@ struct Instance {
     PFN_vkCmdSetViewport vkCmdSetViewport;
     PFN_vkCmdBindDescriptorSets vkCmdBindDescriptorSets;
     PFN_vkCmdCopyBuffer vkCmdCopyBuffer;
-    PFN_vkCmdDrawIndexedIndirect vkCmdDrawIndexedIndirect;
     PFN_vkCmdPushConstants vkCmdPushConstants;
     PFN_vkUnmapMemory vkUnmapMemory;
     PFN_vkEndCommandBuffer vkEndCommandBuffer;
@@ -206,16 +230,16 @@ struct Instance {
     PFN_vkDestroyBuffer vkDestroyBuffer;
     PFN_vkUpdateDescriptorSets vkUpdateDescriptorSets;
     PFN_vkCreateRenderPass vkCreateRenderPass;
-    PFN_vkCmdDraw vkCmdDraw;
     PFN_vkCmdBindVertexBuffers vkCmdBindVertexBuffers;
     PFN_vkAllocateMemory vkAllocateMemory;
     PFN_vkBindBufferMemory vkBindBufferMemory;
     PFN_vkCreatePipelineLayout vkCreatePipelineLayout;
+    PFN_vkCreatePipelineCache vkCreatePipelineCache;
+    PFN_vkGetPipelineCacheData vkGetPipelineCacheData;
     PFN_vkCmdSetScissor vkCmdSetScissor;
     PFN_vkCmdBindPipeline vkCmdBindPipeline;
     PFN_vkCreateGraphicsPipelines vkCreateGraphicsPipelines;
     PFN_vkCreateDescriptorSetLayout vkCreateDescriptorSetLayout;
-    PFN_vkCmdDrawIndexed vkCmdDrawIndexed;
     PFN_vkCmdEndRenderPass vkCmdEndRenderPass;
     PFN_vkCmdPipelineBarrier vkCmdPipelineBarrier;
     PFN_vkCreateDescriptorPool vkCreateDescriptorPool;
@@ -236,6 +260,16 @@ struct Instance {
     PFN_vkCmdCopyImage vkCmdCopyImage;
     PFN_vkCmdBlitImage vkCmdBlitImage;
 
+    PFN_vkCmdDraw vkCmdDraw;
+    PFN_vkCmdDrawIndexed vkCmdDrawIndexed;
+    PFN_vkCmdDrawIndirect vkCmdDrawIndirect;
+    PFN_vkCmdDrawIndexedIndirect vkCmdDrawIndexedIndirect;
+    PFN_vkCmdDrawIndirectCount vkCmdDrawIndirectCount;
+    PFN_vkCmdDrawIndexedIndirectCount vkCmdDrawIndexedIndirectCount;
+    PFN_vkCmdDrawMeshTasksNV vkCmdDrawMeshTasksNV;
+    PFN_vkCmdDrawMeshTasksIndirectNV vkCmdDrawMeshTasksIndirectNV;
+    PFN_vkCmdDrawMeshTasksIndirectCountNV vkCmdDrawMeshTasksIndirectCountNV;
+
     PFN_vkGetPhysicalDeviceSurfaceSupportKHR vkGetPhysicalDeviceSurfaceSupportKHR;
     PFN_vkGetPhysicalDeviceSurfaceFormatsKHR vkGetPhysicalDeviceSurfaceFormatsKHR;
     PFN_vkGetPhysicalDeviceSurfaceCapabilitiesKHR vkGetPhysicalDeviceSurfaceCapabilitiesKHR;
@@ -323,10 +357,12 @@ struct RenderPipeline {
     PyObject_HEAD
     Instance * instance;
     RenderParameters parameters;
+    RenderCommand render_command;
     Buffer * vertex_buffer;
     Buffer * instance_buffer;
     Buffer * index_buffer;
     Buffer * indirect_buffer;
+    Buffer * count_buffer;
     uint32_t binding_count;
     DescriptorBinding * binding_array;
     VkDescriptorSetLayoutBinding * descriptor_binding_array;
@@ -336,8 +372,6 @@ struct RenderPipeline {
     VkPipelineLayout pipeline_layout;
     VkDescriptorPool descriptor_pool;
     VkDescriptorSet descriptor_set;
-    VkShaderModule vertex_shader_module;
-    VkShaderModule fragment_shader_module;
     uint32_t attribute_count;
     VkBuffer * attribute_buffer_array;
     VkDeviceSize * attribute_offset_array;
@@ -358,7 +392,6 @@ struct ComputePipeline {
     VkPipelineLayout pipeline_layout;
     VkDescriptorPool descriptor_pool;
     VkDescriptorSet descriptor_set;
-    VkShaderModule compute_shader_module;
     VkPipeline pipeline;
     PyObject * members;
 };
@@ -450,13 +483,26 @@ T * allocate(uint32_t count) {
     return (T *)PyMem_Malloc(sizeof(T) * count);
 }
 
+inline bool has_key(PyObject * dict, const char * key) {
+    return !!PyDict_GetItemString(dict, key);
+}
+
 PyObject * Buffer_meth_write(Buffer * self, PyObject * arg);
 
 PFN_vkGetInstanceProcAddr get_instance_proc_addr(const char * backend);
+
 void load_library_methods(Instance * instance);
 void load_instance_methods(Instance * instance);
 void load_device_methods(Instance * instance);
 
+PyObject * get_instance_layers(PFN_vkEnumerateInstanceLayerProperties vkEnumerateInstanceLayerProperties);
+PyObject * get_instance_extensions(PFN_vkEnumerateInstanceExtensionProperties vkEnumerateInstanceExtensionProperties);
+PyObject * get_device_extensions(VkPhysicalDevice physical_device, PFN_vkEnumerateDeviceExtensionProperties vkEnumerateDeviceExtensionProperties);
+
+uint32_t load_instance_layers(Instance * instance, const char ** array, PyObject * extra_layers);
+uint32_t load_instance_extensions(Instance * instance, const char ** array, const char * surface);
+uint32_t load_device_extensions(Instance * instance, const char ** array);
+
 void install_debug_messenger(Instance * instance);
 
 int parse_descriptor_binding(Instance * instance, DescriptorBinding * binding, PyObject * obj);
@@ -479,7 +525,7 @@ Memory * get_memory(Instance * instance, PyObject * memory);
 
 VkDeviceSize take_memory(Memory * self, VkMemoryRequirements * requirements);
 
-void allocate_memory(Memory * self);
+void allocate_memory(Memory * self, VkMemoryDedicatedAllocateInfo * dedicated = NULL);
 void free_memory(Memory * self);
 
 Image * new_image(ImageCreateInfo info);
