@@ -11,21 +11,114 @@ Buffer * get_buffer(Instance * instance, PyObject * obj) {
     return NULL;
 }
 
+const int indirect_mesh_task_stride = 8;
+const int indirect_indexed_stride = 20;
+const int indirect_stride = 16;
+
+void render_mesh_task_indirect_count(RenderPipeline * self, VkCommandBuffer command_buffer) {
+    self->instance->vkCmdDrawMeshTasksIndirectCountNV(
+        command_buffer,
+        self->indirect_buffer->buffer,
+        0,
+        self->count_buffer->buffer,
+        0,
+        self->parameters.max_draw_count,
+        indirect_mesh_task_stride
+    );
+}
+
+void render_mesh_task_indirect(RenderPipeline * self, VkCommandBuffer command_buffer) {
+    self->instance->vkCmdDrawMeshTasksIndirectNV(
+        command_buffer,
+        self->indirect_buffer->buffer,
+        self->parameters.indirect_buffer_offset,
+        self->parameters.instance_count,
+        indirect_mesh_task_stride
+    );
+}
+
+void render_mesh_task(RenderPipeline * self, VkCommandBuffer command_buffer) {
+    self->instance->vkCmdDrawMeshTasksNV(command_buffer, self->parameters.instance_count, 0);
+}
+
+void render_indirect_indexed_count(RenderPipeline * self, VkCommandBuffer command_buffer) {
+    self->instance->vkCmdDrawIndexedIndirectCount(
+        command_buffer,
+        self->indirect_buffer->buffer,
+        self->parameters.indirect_buffer_offset,
+        self->count_buffer->buffer,
+        self->parameters.count_buffer_offset,
+        self->parameters.max_draw_count,
+        indirect_indexed_stride
+    );
+}
+
+void render_indirect_count(RenderPipeline * self, VkCommandBuffer command_buffer) {
+    self->instance->vkCmdDrawIndirectCount(
+        command_buffer,
+        self->indirect_buffer->buffer,
+        self->parameters.indirect_buffer_offset,
+        self->count_buffer->buffer,
+        self->parameters.count_buffer_offset,
+        self->parameters.max_draw_count,
+        indirect_stride
+    );
+}
+
+void render_indirect_indexed(RenderPipeline * self, VkCommandBuffer command_buffer) {
+    self->instance->vkCmdDrawIndexedIndirect(
+        command_buffer,
+        self->indirect_buffer->buffer,
+        self->parameters.indirect_buffer_offset,
+        self->parameters.indirect_count,
+        indirect_indexed_stride
+    );
+}
+
+void render_indirect(RenderPipeline * self, VkCommandBuffer command_buffer) {
+    self->instance->vkCmdDrawIndirect(
+        command_buffer,
+        self->indirect_buffer->buffer,
+        self->parameters.indirect_buffer_offset,
+        self->parameters.indirect_count,
+        indirect_stride
+    );
+}
+
+void render_indexed(RenderPipeline * self, VkCommandBuffer command_buffer) {
+    self->instance->vkCmdDrawIndexed(command_buffer, self->parameters.index_count, self->parameters.instance_count, 0, 0, 0);
+}
+
+void render_simple(RenderPipeline * self, VkCommandBuffer command_buffer) {
+    self->instance->vkCmdDraw(command_buffer, self->parameters.vertex_count, self->parameters.instance_count, 0, 0);
+}
+
 RenderPipeline * Framebuffer_meth_render(Framebuffer * self, PyObject * vargs, PyObject * kwargs) {
     static char * keywords[] = {
         "vertex_shader",
         "fragment_shader",
+        "task_shader",
+        "mesh_shader",
         "vertex_format",
         "instance_format",
         "vertex_count",
         "instance_count",
         "index_count",
         "indirect_count",
+        "max_draw_count",
         "vertex_buffer",
         "instance_buffer",
         "index_buffer",
         "indirect_buffer",
+        "count_buffer",
+        "vertex_buffer_offset",
+        "instance_buffer_offset",
+        "index_buffer_offset",
+        "indirect_buffer_offset",
+        "count_buffer_offset",
         "topology",
+        "depth_test",
+        "depth_write",
         "bindings",
         "memory",
         NULL,
@@ -34,17 +127,28 @@ RenderPipeline * Framebuffer_meth_render(Framebuffer * self, PyObject * vargs, P
     struct {
         PyObject * vertex_shader = Py_None;
         PyObject * fragment_shader = Py_None;
+        PyObject * task_shader = Py_None;
+        PyObject * mesh_shader = Py_None;
         PyObject * vertex_format;
         PyObject * instance_format;
         uint32_t vertex_count = 0;
         uint32_t instance_count = 1;
         uint32_t index_count = 0;
         uint32_t indirect_count = 0;
+        uint32_t max_draw_count = 0;
         PyObject * vertex_buffer = Py_None;
         PyObject * instance_buffer = Py_None;
         PyObject * index_buffer = Py_None;
         PyObject * indirect_buffer = Py_None;
+        PyObject * count_buffer = Py_None;
+        VkDeviceSize vertex_buffer_offset = 0;
+        VkDeviceSize instance_buffer_offset = 0;
+        VkDeviceSize index_buffer_offset = 0;
+        VkDeviceSize indirect_buffer_offset = 0;
+        VkDeviceSize count_buffer_offset = 0;
         PyObject * topology;
+        VkBool32 depth_test = true;
+        VkBool32 depth_write = true;
         PyObject * bindings;
         PyObject * memory = Py_None;
     } args;
@@ -57,23 +161,36 @@ RenderPipeline * Framebuffer_meth_render(Framebuffer * self, PyObject * vargs, P
     int args_ok = PyArg_ParseTupleAndKeywords(
         vargs,
         kwargs,
-        "|$O!O!OOIIIIOOOOOOO",
+        "|$O!O!O!O!OOIIIIIOOOOOKKKKKOppOO",
         keywords,
         &PyBytes_Type,
         &args.vertex_shader,
         &PyBytes_Type,
         &args.fragment_shader,
+        &PyBytes_Type,
+        &args.task_shader,
+        &PyBytes_Type,
+        &args.mesh_shader,
         &args.vertex_format,
         &args.instance_format,
         &args.vertex_count,
         &args.instance_count,
         &args.index_count,
         &args.indirect_count,
+        &args.max_draw_count,
         &args.vertex_buffer,
         &args.instance_buffer,
         &args.index_buffer,
         &args.indirect_buffer,
+        &args.count_buffer,
+        &args.vertex_buffer_offset,
+        &args.instance_buffer_offset,
+        &args.index_buffer_offset,
+        &args.indirect_buffer_offset,
+        &args.count_buffer_offset,
         &args.topology,
+        &args.depth_test,
+        &args.depth_write,
         &args.bindings,
         &args.memory
     );
@@ -89,6 +206,20 @@ RenderPipeline * Framebuffer_meth_render(Framebuffer * self, PyObject * vargs, P
     res->instance = self->instance;
     res->members = PyDict_New();
 
+    res->parameters = {
+        true,
+        args.vertex_count,
+        args.instance_count,
+        args.index_count,
+        args.indirect_count,
+        args.max_draw_count,
+        args.vertex_buffer_offset,
+        args.instance_buffer_offset,
+        args.index_buffer_offset,
+        args.indirect_buffer_offset,
+        args.count_buffer_offset,
+    };
+
     PyObject * vertex_format = PyUnicode_Split(args.vertex_format, NULL, -1);
     PyObject * instance_format = PyUnicode_Split(args.instance_format, NULL, -1);
 
@@ -127,17 +258,8 @@ RenderPipeline * Framebuffer_meth_render(Framebuffer * self, PyObject * vargs, P
         binding_array[i] = {i, istride, VK_VERTEX_INPUT_RATE_INSTANCE};
     }
 
-    res->parameters = {
-        true,
-        args.vertex_count,
-        args.instance_count,
-        args.index_count,
-        args.indirect_count,
-    };
-
     VkBool32 short_index = false;
     uint32_t index_size = short_index ? 2 : 4;
-    uint32_t indirect_size = args.index_count ? 20 : 16;
 
     res->binding_count = (uint32_t)PyList_Size(args.bindings);
     res->binding_array = (DescriptorBinding *)PyMem_Malloc(sizeof(DescriptorBinding) * PyList_Size(args.bindings));
@@ -152,6 +274,17 @@ RenderPipeline * Framebuffer_meth_render(Framebuffer * self, PyObject * vargs, P
     res->instance_buffer = get_buffer(self->instance, args.instance_buffer);
     res->index_buffer = get_buffer(self->instance, args.index_buffer);
     res->indirect_buffer = get_buffer(self->instance, args.indirect_buffer);
+    res->count_buffer = get_buffer(self->instance, args.count_buffer);
+
+    uint32_t indirect_size = indirect_stride;
+
+    if (args.index_count || res->index_buffer) {
+        indirect_size = indirect_indexed_stride;
+    }
+
+    if (args.mesh_shader != Py_None) {
+        indirect_size = indirect_mesh_task_stride;
+    }
 
     if (PyErr_Occurred()) {
         return NULL;
@@ -209,6 +342,10 @@ RenderPipeline * Framebuffer_meth_render(Framebuffer * self, PyObject * vargs, P
         PyDict_SetItemString(res->members, "indirect_buffer", (PyObject *)res->indirect_buffer);
     }
 
+    if (res->count_buffer) {
+        PyDict_SetItemString(res->members, "count_buffer", (PyObject *)res->count_buffer);
+    }
+
     for (uint32_t i = 0; i < res->binding_count; ++i) {
         create_descriptor_binding_objects(self->instance, &res->binding_array[i], memory);
     }
@@ -241,12 +378,12 @@ RenderPipeline * Framebuffer_meth_render(Framebuffer * self, PyObject * vargs, P
 
     for (uint32_t i = 0; i < vertex_attribute_count; ++i) {
         res->attribute_buffer_array[i] = res->vertex_buffer->buffer;
-        res->attribute_offset_array[i] = 0;
+        res->attribute_offset_array[i] = args.vertex_buffer_offset;
     }
 
     for (uint32_t i = vertex_attribute_count; i < attribute_count; ++i) {
         res->attribute_buffer_array[i] = res->instance_buffer->buffer;
-        res->attribute_offset_array[i] = 0;
+        res->attribute_offset_array[i] = args.instance_buffer_offset;
     }
 
     res->descriptor_binding_array = (VkDescriptorSetLayoutBinding *)PyMem_Malloc(sizeof(VkDescriptorSetLayoutBinding) * res->binding_count);
@@ -329,46 +466,77 @@ RenderPipeline * Framebuffer_meth_render(Framebuffer * self, PyObject * vargs, P
 
     self->instance->vkCreatePipelineLayout(self->instance->device, &pipeline_layout_create_info, NULL, &res->pipeline_layout);
 
-    VkShaderModuleCreateInfo vertex_shader_module_create_info = {
-        VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO,
-        NULL,
-        0,
-        (VkDeviceSize)PyBytes_Size(args.vertex_shader),
-        (uint32_t *)PyBytes_AsString(args.vertex_shader),
-    };
+    uint32_t pipeline_shader_stage_count = 0;
+    VkPipelineShaderStageCreateInfo pipeline_shader_stage_array[8] = {};
 
-    self->instance->vkCreateShaderModule(self->instance->device, &vertex_shader_module_create_info, NULL, &res->vertex_shader_module);
+    if (args.vertex_shader != Py_None) {
+        VkShaderModuleCreateInfo vertex_shader_module_create_info = {
+            VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO,
+            NULL,
+            0,
+            (VkDeviceSize)PyBytes_Size(args.vertex_shader),
+            (uint32_t *)PyBytes_AsString(args.vertex_shader),
+        };
 
-    VkShaderModuleCreateInfo fragment_shader_module_create_info = {
-        VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO,
-        NULL,
-        0,
-        (VkDeviceSize)PyBytes_Size(args.fragment_shader),
-        (uint32_t *)PyBytes_AsString(args.fragment_shader),
-    };
+        VkShaderModule vertex_shader_module = NULL;
+        self->instance->vkCreateShaderModule(self->instance->device, &vertex_shader_module_create_info, NULL, &vertex_shader_module);
 
-    self->instance->vkCreateShaderModule(self->instance->device, &fragment_shader_module_create_info, NULL, &res->fragment_shader_module);
-
-    VkPipelineShaderStageCreateInfo pipeline_shader_stage_array[] = {
-        {
+        pipeline_shader_stage_array[pipeline_shader_stage_count++] = {
             VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO,
             NULL,
             0,
             VK_SHADER_STAGE_VERTEX_BIT,
-            res->vertex_shader_module,
+            vertex_shader_module,
             "main",
             NULL,
-        },
-        {
+        };
+    }
+
+    if (args.fragment_shader != Py_None) {
+        VkShaderModuleCreateInfo fragment_shader_module_create_info = {
+            VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO,
+            NULL,
+            0,
+            (VkDeviceSize)PyBytes_Size(args.fragment_shader),
+            (uint32_t *)PyBytes_AsString(args.fragment_shader),
+        };
+
+        VkShaderModule fragment_shader_module = NULL;
+        self->instance->vkCreateShaderModule(self->instance->device, &fragment_shader_module_create_info, NULL, &fragment_shader_module);
+
+        pipeline_shader_stage_array[pipeline_shader_stage_count++] = {
             VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO,
             NULL,
             0,
             VK_SHADER_STAGE_FRAGMENT_BIT,
-            res->fragment_shader_module,
+            fragment_shader_module,
             "main",
             NULL,
-        },
-    };
+        };
+    }
+
+    if (args.mesh_shader != Py_None) {
+        VkShaderModuleCreateInfo mesh_shader_module_create_info = {
+            VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO,
+            NULL,
+            0,
+            (VkDeviceSize)PyBytes_Size(args.mesh_shader),
+            (uint32_t *)PyBytes_AsString(args.mesh_shader),
+        };
+
+        VkShaderModule mesh_shader_module = NULL;
+        self->instance->vkCreateShaderModule(self->instance->device, &mesh_shader_module_create_info, NULL, &mesh_shader_module);
+
+        pipeline_shader_stage_array[pipeline_shader_stage_count++] = {
+            VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO,
+            NULL,
+            0,
+            VK_SHADER_STAGE_MESH_BIT_NV,
+            mesh_shader_module,
+            "main",
+            NULL,
+        };
+    }
 
     VkPipelineVertexInputStateCreateInfo pipeline_vertex_input_state = {
         VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO,
@@ -430,8 +598,8 @@ RenderPipeline * Framebuffer_meth_render(Framebuffer * self, PyObject * vargs, P
         VK_STRUCTURE_TYPE_PIPELINE_DEPTH_STENCIL_STATE_CREATE_INFO,
         NULL,
         0,
-        true,
-        true,
+        args.depth_test,
+        args.depth_write,
         VK_COMPARE_OP_LESS,
         false,
         false,
@@ -478,11 +646,11 @@ RenderPipeline * Framebuffer_meth_render(Framebuffer * self, PyObject * vargs, P
         dynamic_state_array,
     };
 
-    VkGraphicsPipelineCreateInfo graphics_pipeline_create_info = {
+    VkGraphicsPipelineCreateInfo pipeline_create_info = {
         VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO,
         NULL,
         0,
-        2,
+        pipeline_shader_stage_count,
         pipeline_shader_stage_array,
         &pipeline_vertex_input_state,
         &pipeline_input_assembly_state,
@@ -500,7 +668,35 @@ RenderPipeline * Framebuffer_meth_render(Framebuffer * self, PyObject * vargs, P
         0,
     };
 
-    self->instance->vkCreateGraphicsPipelines(self->instance->device, NULL, 1, &graphics_pipeline_create_info, NULL, &res->pipeline);
+    self->instance->vkCreateGraphicsPipelines(self->instance->device, self->instance->pipeline_cache, 1, &pipeline_create_info, NULL, &res->pipeline);
+
+    for (uint32_t i = 0; i < pipeline_shader_stage_count; ++i) {
+        self->instance->vkDestroyShaderModule(self->instance->device, pipeline_shader_stage_array[i].module, NULL);
+    }
+
+    if (args.mesh_shader != Py_None) {
+        if (res->indirect_buffer && res->count_buffer) {
+            res->render_command = render_mesh_task_indirect_count;
+        } else if (res->indirect_buffer) {
+            res->render_command = render_mesh_task_indirect;
+        } else {
+            res->render_command = render_mesh_task;
+        }
+    } else {
+        if (res->indirect_buffer && res->index_buffer && res->count_buffer) {
+            res->render_command = render_indirect_indexed_count;
+        } else if (res->indirect_buffer && res->count_buffer) {
+            res->render_command = render_indirect_count;
+        } else if (res->indirect_buffer && res->index_buffer) {
+            res->render_command = render_indirect_indexed;
+        } else if (res->indirect_buffer) {
+            res->render_command = render_indirect;
+        } else if (res->index_buffer) {
+            res->render_command = render_indexed;
+        } else {
+            res->render_command = render_simple;
+        }
+    }
 
     for (uint32_t i = 0; i < res->binding_count; ++i) {
         if (res->binding_array[i].name) {
@@ -600,20 +796,12 @@ void execute_render_pipeline(RenderPipeline * self, VkCommandBuffer command_buff
         self->instance->vkCmdBindIndexBuffer(
             command_buffer,
             self->index_buffer->buffer,
-            0,
+            self->parameters.index_buffer_offset,
             VK_INDEX_TYPE_UINT32
         );
     }
 
-    if (self->indirect_buffer && self->index_buffer) {
-        self->instance->vkCmdDrawIndexedIndirect(command_buffer, self->indirect_buffer->buffer, 0, self->parameters.indirect_count, 20);
-    } else if (self->indirect_buffer) {
-        self->instance->vkCmdDrawIndirect(command_buffer, self->indirect_buffer->buffer, 0, self->parameters.indirect_count, 16);
-    } else if (self->index_buffer) {
-        self->instance->vkCmdDrawIndexed(command_buffer, self->parameters.index_count, self->parameters.instance_count, 0, 0, 0);
-    } else {
-        self->instance->vkCmdDraw(command_buffer, self->parameters.vertex_count, self->parameters.instance_count, 0, 0);
-    }
+    self->render_command(self, command_buffer);
 }
 
 PyObject * RenderPipeline_subscript(RenderPipeline * self, PyObject * key) {
